# kytten/scrollbar.py
# Copyrighted (C) 2009 by Conrad "Lynx" Wong

import pyglet
from widgets import Control

class HScrollbar(Control):
    """
    A horizontal scrollbar.  Position is measured from 0.0 to 1.0, and bar
    size is set as a percentage of the maximum.
    """
    IMAGE_LEFT = ['hscrollbar', 'left']
    IMAGE_SPACE = ['hscrollbar', 'space']
    IMAGE_BAR = ['hscrollbar', 'bar']
    IMAGE_RIGHT = ['hscrollbar', 'right']
    IMAGE_LEFTMAX = ['hscrollbar', 'leftmax']
    IMAGE_RIGHTMAX = ['hscrollbar', 'rightmax']

    def __init__(self, width):
        """
        Creates a new scrollbar.

        @param width Width of the area for which we are a scrollbar
        """
        Control.__init__(self, width=width, height=0)
        self.__init2__(width)

    def __init2__(self, width):
        """
        HScrollbar and VScrollbar share similiar data structures, which this
        function initializes.

        @param width Width of the area for which we are a scrollbar
        """
        self.left = None
        self.space = None
        self.right = None
        self.bar = None
        self.pos = 0.0
        self.bar_width = 0.5
        self.is_dragging = False
        self.is_scrolling = False
        self.scroll_delta = 0

    def _get_left_region(self):
        """
        Return area of the left button (x, y, width, height)
        """
        if self.left is not None:
            return self.x, self.y, self.left.width, self.height
        else:
            return self.x, self.y, 0, 0

    def _get_right_region(self):
        """
        Return area of the right button (x, y, width, height)
        """
        if self.right is not None:
            return (self.x + self.width - self.right.width, self.y,
                    self.right.width, self.height)
        else:
            return self.x + self.width, self.y, 0, 0

    def _get_space_region(self):
        """
        Return area of the space in which the bar moves
        (x, y, width, height)
        """
        if self.left is not None and self.right is not None:
            return (self.x + self.left.width, self.y,
                    self.width - self.left.width - self.right.width,
                    self.height)
        else:
            return self.x, self.y, self.width, self.height

    def _get_bar_region(self):
        """
        Return area of the bar within the scrollbar (x, y, width, height)
        """
        if self.left is not None and self.right is not None:
            left_width = self.left.width
            right_width = self.right.width
        else:
            left_width = right_width = 0
        self.pos = max(min(self.pos, 1.0 - self.bar_width), 0.0)
        space_width = self.width - left_width - right_width
        return (int(self.x + left_width + self.pos * space_width),
                self.y,
                int(self.bar_width * space_width),
                self.height)

    def delete(self):
        """
        Delete all graphic elements used by the scrollbar
        """
        if self.left is not None:
            self.left.delete()
            self.left = None
        if self.space is not None:
            self.space.delete()
            self.space = None
        if self.bar is not None:
            self.bar.delete()
            self.bar = None
        if self.right is not None:
            self.right.delete()
            self.right = None

    def drag_bar(self, dx, dy):
        """
        Drag the bar, keeping it within limits

        @param dx Delta X
        @param dy Delta Y
        """
        _, _, space_width, space_height = self._get_space_region()
        _, _, bar_width, bar_height = self._get_bar_region()
        self.pos = min(max(self.pos + float(dx) / space_width, 0.0),
                       1.0 - float(bar_width)/space_width)

    def ensure_visible(self, left, right, max_width):
        """
        Ensure that the area of space between left and right is completely
        visible.

        @param left Left end of the space
        @param right Right end of the space
        @param max_width Maximum width of space
        """
        pos_x = self.pos * max_width
        pos_width = self.bar_width * max_width
        if pos_x <= left and pos_x + pos_width > right:
            return  # We're fine
        elif pos_x > left:
            self.pos = pos_x / max_width  # Shift to the left
        elif pos_x + pos_width < right:
            self.pos = (right - pos_width) / max_width  # Shift to the right
        self.pos = min(max(self.pos, 0.0), 1.0 - self.bar_width)
        self.delete()
        self.saved_dialog.set_needs_layout()

    def get(self, width, max_width):
        """
        Returns the position of the bar, in pixels from the controlled area's
        left edge
        """
        return int(self.pos * max_width)

    def layout(self, x, y):
        """
        Lays out the scrollbar components

        @param x X coordinate of lower left corner
        @param y Y coordinate of lower left corner
        """
        self.x, self.y = x, y
        if self.left is not None:
            self.left.update(*self._get_left_region())
        if self.right is not None:
            self.right.update(*self._get_right_region())
        if self.space is not None:
            self.space.update(*self._get_space_region())
        if self.bar is not None:
            self.bar.update(*self._get_bar_region())

    def on_gain_focus(self):
        if self.saved_dialog is not None:
            self.saved_dialog.set_wheel_target(self)

    def on_lose_focus(self):
        if self.saved_dialog is not None:
            self.saved_dialog.set_wheel_target(None)

    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        """
        We drag the bar only if the user had previously clicked on the bar

        @param x X coordinate of mouse
        @param y Y coordinate of mouse
        @param dx Delta X
        @param dy Delta Y
        @param buttons Buttons held while dragging
        @param modifiers Modifiers to apply to buttons
        """
        if self.is_dragging:
            self.drag_bar(dx, dy)
            self.delete()
            self.saved_dialog.set_needs_layout()
            return pyglet.event.EVENT_HANDLED

    def on_mouse_press(self, x, y, button, modifiers):
        """
        If the mouse press falls within the space, move the bar over to the
        mouse.  Otherwise, activate scrolling.

        @param x X coordinate of mouse
        @param y Y coordinate of mouse
        @param button Button being pressed
        @param modifiers Modifiers to apply to button
        """
        space_x, space_y, space_width, space_height = self._get_space_region()
        if x >= space_x and x < space_x + space_width and \
           y >= space_y and y < space_y + space_height:
            self.set_bar_pos(x, y)
            self.is_dragging = True
            self.delete()
            self.saved_dialog.set_needs_layout()
        else:
            left_x, left_y, left_width, left_height = self._get_left_region()
            if x >= left_x and x < left_x + left_width and \
               y >= left_y and y < left_y + left_height:
                self.is_scrolling = True
                self.scroll_delta = -1
            else:
                right_x, right_y, right_width, right_height = \
                       self._get_right_region()
                if x >= right_x and x < right_x + right_width and \
                   y >= right_y and y < right_y + right_height:
                    self.is_scrolling = True
                    self.scroll_delta = 1

    def on_mouse_release(self, x, y, button, modifiers):
        """
        Cancels dragging or scrolling

        @param x X coordinate of mouse
        @param y Y coordinate of mouse
        @param button Button being released
        @param modifiers Modifiers to apply to button
        """
        self.is_dragging = False
        self.is_scrolling = False
        self.scroll_delta = 0

    def on_mouse_scroll(self, x, y, scroll_x, scroll_y):
        """
        Mousewheel was moved some number of clicks

        @param x X coordinate of mouse
        @param y Y coordinate of mouse
        @param scroll_x Number of clicks horizontally mouse was moved
        @param scroll_y Number of clicks vertically mouse was moved
        """
        self.drag_bar(scroll_y * 10, 0)
        self.delete()
        self.saved_dialog.set_needs_layout()

    def on_update(self, dt):
        """
        When scrolling, we increment our position each update

        @param dt Time delta, in seconds
        """
        if self.is_scrolling:
            self.drag_bar(self.scroll_delta * 50.0 * dt, 0)
            self.saved_dialog.set_needs_layout()

    def set(self, width, max_width):
        """
        Sets the width of the scrollbar

        @param width Width the scrollbar occupies
        @param max_width Maximum width of the scrollable area
        """
        self.width = width
        self.bar_width = max(float(width) / max_width, 0.0)
        self.pos = min(self.pos, 1.0 - self.bar_width)

    def set_bar_pos(self, x, y):
        """
        When the mouse is pressed within scrollbar space, move the bar over
        underneath it.

        @param x X coordinate of mouse press
        @param y Y coordinate of mouse press
        """
        space_x, space_y, space_width, space_height = self._get_space_region()
        bar_x, bar_y, bar_width, bar_height = self._get_bar_region()
        if x < bar_x:
            self.pos = float(x - space_x) / space_width
        elif x > bar_x + bar_width:
            max_bar_x = space_width - bar_width
            x -= bar_width
            self.pos = float(min(max_bar_x, x - space_x)) / space_width
        if self.bar is not None:
            self.bar.update(*self._get_bar_region())

    def size(self, dialog):
        """
        Creates scrollbar components.
        """
        if dialog is None:
            return
        Control.size(self, dialog)
        dialog.set_wheel_hint(self)
        if self.left is None:
            if self.pos > 0.0:
                path = self.IMAGE_LEFT
            else:
                path = self.IMAGE_LEFTMAX
            self.left = dialog.theme[path]['image'].generate(
                dialog.theme[path]['gui_color'],
                dialog.batch, dialog.fg_group)

            # Left button is our basis for minimum dimension
            self.width, self.height = self.left.width, self.left.height
        if self.space is None:
            path = self.IMAGE_SPACE
            self.space = dialog.theme[path]['image'].generate(
                dialog.theme[path]['gui_color'],
                dialog.batch, dialog.fg_group)
        if self.bar is None:
            path = self.IMAGE_BAR
            self.bar = dialog.theme[path]['image'].generate(
                dialog.theme[path]['gui_color'],
                dialog.batch, dialog.fg_group)
        if self.right is None:
            if self.pos < 1.0 - self.bar_width:
                path = self.IMAGE_RIGHT
            else:
                path = self.IMAGE_RIGHTMAX
            self.right = dialog.theme[path]['image'].generate(
                dialog.theme[path]['gui_color'],
                dialog.batch, dialog.fg_group)

class VScrollbar(HScrollbar):
    """
    A vertical scrollbar.  Position is measured from 0.0 to 1.0, and bar size
    is set as a percentage of the maximum.  Note that left is top, and
    right is bottom, from the viewpoint of the VScrollbar.
    """
    IMAGE_LEFT = ['vscrollbar', 'up']
    IMAGE_SPACE = ['vscrollbar', 'space']
    IMAGE_BAR = ['vscrollbar', 'bar']
    IMAGE_RIGHT = ['vscrollbar', 'down']
    IMAGE_LEFTMAX = ['vscrollbar', 'upmax']
    IMAGE_RIGHTMAX = ['vscrollbar', 'downmax']

    def __init__(self, height):
        """
        Creates a new scrollbar.  At the outset, we are presented with maximum
        height and the templates to use.

        @param height Height of the area for which we are a scrollbar
        """
        Control.__init__(self, width=0, height=height)
        self.__init2__(height)

    def _get_left_region(self):
        """Returns the area occupied by the up button
        (x, y, width, height)"""
        if self.left is not None:
            return (self.x, self.y + self.height - self.left.height,
                    self.width, self.left.height)
        else:
            return self.x, self.y, 0, 0

    def _get_right_region(self):
        """Returns the area occupied by the down button
        (x, y, width, height)"""
        if self.right is not None:
            return self.x, self.y, self.width, self.right.height
        else:
            return self.x, self.y, 0, 0

    def _get_space_region(self):
        """Returns the area occupied by the space between up and down
        buttons (x, y, width, height)"""
        if self.left is not None and self.right is not None:
            return (self.x,
                    self.y + self.right.height,
                    self.width,
                    self.height - self.left.width - self.right.width)
        else:
            return self.x, self.y, self.width, self.height

    def _get_bar_region(self):
        """Returns the area occupied by the bar (x, y, width, height)"""
        if self.left is not None and self.right is not None:
            left_height = self.left.height
            right_height = self.right.height
        else:
            left_height = right_height = 0
        self.pos = max(min(self.pos, 1.0 - self.bar_width), 0.0)
        space_height = self.height - left_height - right_height
        top = self.y + self.height - left_height
        return (self.x, int(top - (self.pos + self.bar_width) * space_height),
                self.width, int(self.bar_width * space_height))

    def drag_bar(self, dx, dy):
        """Handles dragging the bar.

        @param dx Delta X
        @param dy Delta Y
        """
        _, _, space_width, space_height = self._get_space_region()
        _, _, bar_width, bar_height = self._get_bar_region()
        self.pos = min(max(self.pos - float(dy) / space_height, 0.0),
                       1.0 - float(bar_height)/space_height)

    def ensure_visible(self, top, bottom, max_height):
        """
        Ensure that the area of space between top and bottom is completely
        visible.

        @param top Top end of the space
        @param bottom Bottom end of the space
        @param max_height Maximum height of space
        """
        bar_top = (1.0 - self.pos) * max_height
        bar_bottom = bar_top - self.bar_width * max_height
        if bar_top > top and bar_bottom <= bottom:
            return  # We're fine
        elif bar_top < top:
            # Shift upward
            self.pos = 1.0 - float(top) / max_height
        elif bar_bottom > bottom:
            # Shift downward
            self.pos = 1.0 - float(bottom) / max_height - self.bar_width
        self.pos = min(max(self.pos, 0.0), 1.0 - self.bar_width)
        self.delete()
        self.saved_dialog.set_needs_layout()

    def on_mouse_scroll(self, x, y, scroll_x, scroll_y):
        """
        Mousewheel was moved some number of clicks

        @param x X coordinate of mouse
        @param y Y coordinate of mouse
        @param scroll_x Number of clicks horizontally mouse was moved
        @param scroll_y Number of clicks vertically mouse was moved
        """
        self.drag_bar(0, scroll_y * 10)
        self.delete()
        self.saved_dialog.set_needs_layout()

    def on_update(self, dt):
        """
        When scrolling, we increment our position each update

        @param dialog Dialog in which we're contained
        @param dt Time delta, in seconds
        """
        if self.is_scrolling:
            self.drag_bar(0, -self.scroll_delta * 50.0 * dt)
            self.saved_dialog.set_needs_layout()

    def set(self, height, max_height):
        """Sets the new height of the scrollbar, and the height of
        the bar relative to the scrollable area.

        @param height Scrollable region height
        @param max_height Maximum scrollable height
        """
        self.height = height
        self.bar_width = max(float(height) / max_height, 0.0)
        self.pos = min(self.pos, 1.0 - self.bar_width)

    def set_bar_pos(self, x, y):
        """Sets the scrollbar position.  Moves the scrollbar to intercept
        the mouse if it is not already in place."""
        space_x, space_y, space_width, space_height = self._get_space_region()
        bar_x, bar_y, bar_width, bar_height = self._get_bar_region()
        top = space_y + space_height
        if y > bar_y + bar_height:
            self.pos = float(top - y) / space_height
        elif y < bar_y:
            y += bar_height
            max_bar_y = space_height - bar_height
            self.pos = float(min(max_bar_y, top - y)) / space_height
        if self.bar is not None:
            self.bar.update(*self._get_bar_region())




# Copyright (c) 2009 Conrad "Lynx" Wong
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#  * Neither the name of DarkCoda nor the names of its
#    contributors may be used to endorse or promote products
#    derived from this software without specific prior written
#    permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""kytten - a skinnable, easily constructed GUI toolkit for pyglet

Inspired by simplui (Tristam MacDonald) and many other GUI projects.
Thanks to Gary Herron and Steve Johnson for debugging assistance!
"""

# GUI public constants

from layout import VALIGN_TOP, VALIGN_CENTER, VALIGN_BOTTOM
from layout import HALIGN_LEFT, HALIGN_CENTER, HALIGN_RIGHT
from layout import ANCHOR_TOP_LEFT, ANCHOR_TOP, ANCHOR_TOP_RIGHT, \
                   ANCHOR_LEFT, ANCHOR_CENTER, ANCHOR_RIGHT, \
                   ANCHOR_BOTTOM_LEFT, ANCHOR_BOTTOM, ANCHOR_BOTTOM_RIGHT

# GUI public classes

from button import Button
from checkbox import Checkbox
from dialog import Dialog, PopupMessage, PopupConfirm
from document import Document
from file_dialogs import FileLoadDialog, FileSaveDialog, DirectorySelectDialog
from frame import Frame, TitleFrame, Wrapper, SectionHeader, FoldingSection
from layout import GridLayout, HorizontalLayout, VerticalLayout, FreeLayout
from menu import Menu, Dropdown
from scrollable import Scrollable
from slider import Slider
from text_input import Input
from theme import Theme
from widgets import Widget, Spacer, Label



# color_selector.py
# Copyrighted (C) 2009 by Conrad "Lynx" Wong

import math
import pyglet
from pyglet import gl

import kytten
from kytten.button import Button
from kytten.dialog import Dialog
from kytten.frame import Frame
from kytten.layout import ANCHOR_CENTER, VerticalLayout, HorizontalLayout
from kytten.slider import Slider
from kytten.text_input import Input
from kytten.widgets import Control, Label

class ColorWheel(Control):
    """
    Depicts a circle which can be used to select an RGB color.
    """
    path = ['colorselector', 'crosshair']

    RADIUS = 164
    INNER_RADIUS = 132
    BORDER = 4
    VECTORS = []
    NUM_SEGMENTS = int(2 * math.pi * (RADIUS + BORDER) / 24) * 6
    for x in xrange(0, NUM_SEGMENTS):
	angle = float(x) * 2.0 * math.pi / float(NUM_SEGMENTS)
	VECTORS.append((math.cos(angle), math.sin(angle)))
    NUM_ARC_SEGMENTS = NUM_SEGMENTS / 6
    COLORS = []
    for x in xrange(0, NUM_ARC_SEGMENTS):
	COLORS.append([255, x * 255 / NUM_ARC_SEGMENTS, 0])
    for x in xrange(0, NUM_ARC_SEGMENTS):
	COLORS.append([255 - (x * 255 / NUM_ARC_SEGMENTS), 255, 0])
    for x in xrange(0, NUM_ARC_SEGMENTS):
	COLORS.append([0, 255, x * 255 / NUM_ARC_SEGMENTS])
    for x in xrange(0, NUM_ARC_SEGMENTS):
	COLORS.append([0, 255 - (x * 255 / NUM_ARC_SEGMENTS), 255])
    for x in xrange(0, NUM_ARC_SEGMENTS):
	COLORS.append([x * 255 / NUM_ARC_SEGMENTS, 0, 255])
    for x in xrange(0, NUM_ARC_SEGMENTS):
	COLORS.append([255, 0, 255 - (x * 255 / NUM_ARC_SEGMENTS)])

    def __init__(self, color, id=None, on_select=None):
	Control.__init__(self, id=id)
	self.alpha = color[3]
	self.tip_color = [0, 0, 0]
	self.on_select = on_select
	self.circle_vlist = None
	self.inner_circle_vlist = None
	self.inner_circle_bg_vlist = None
	self.colors_vlist = None
	self.triangle_vlist = None
	self.crosshair = None
	self.tri_angle = 0.0
	self.point = (0, 0)
	self.is_turning = False
	self.is_dragging = False
	self.center_x = 0
	self.center_y = 0
	self.pointer_i = 1.0
	self.pointer_j = 0.0
	self.vector_i = [0, 1]
	self.vector_j = [1, 0]
	self.point_a = [0, 0]

	self.set_color(color)

    def _get_vertices(self, center_x, center_y, inner_radius, outer_radius):
	vertices = []
	lastVec = self.VECTORS[-1]
	for xVec, yVec in self.VECTORS:
	    lastXVec, lastYVec = lastVec
	    vertices.extend([center_x + lastXVec * inner_radius,
			     center_y + lastYVec * inner_radius,
			     center_x + xVec * inner_radius,
			     center_y + yVec * inner_radius,
			     center_x + xVec * outer_radius,
			     center_y + yVec * outer_radius,
			     center_x + lastXVec * outer_radius,
			     center_y + lastYVec * outer_radius])
	    lastVec = (xVec, yVec)

	return vertices

    def _get_circle_vlist_vertices(self):
	radius = self.RADIUS + self.BORDER
	innerRadius = self.INNER_RADIUS - self.BORDER
	return self._get_vertices(self.center_x, self.center_y,
				  innerRadius, radius)

    def _get_inner_circle_vlist_vertices(self):
	center_x = self.center_x
	center_y = self.center_y
	inner_radius = self.INNER_RADIUS - self.BORDER
	vertices = []
	lastVec = self.VECTORS[-1]
	for xVec, yVec in self.VECTORS:
	    lastXVec, lastYVec = lastVec
	    vertices.extend([center_x, center_y,
			     center_x + lastXVec * inner_radius,
			     center_y + lastYVec * inner_radius,
			     center_x + xVec * inner_radius,
			     center_y + yVec * inner_radius])
	    lastVec = (xVec, yVec)
	return vertices

    def _get_inner_circle_bg_vlist_colors(self):
	num_segments = self.NUM_SEGMENTS / 6
	return [255, 255, 255, 255] * (3 * num_segments) + \
	       [0, 0, 0, 255] * (3 * num_segments) + \
	       [255, 255, 255, 255] * (3 * num_segments) + \
	       [0, 0, 0, 255] * (3 * num_segments) + \
	       [255, 255, 255, 255] * (3 * num_segments) + \
	       [0, 0, 0, 255] * (3 * num_segments)

    def _get_colors_vlist_vertices(self):
	radius = self.RADIUS
	innerRadius = self.INNER_RADIUS
	return self._get_vertices(self.center_x, self.center_y,
				  innerRadius, radius)

    def _get_colors_vlist_colors(self):
	colors = []
	lastColor = self.COLORS[-1]
	for color in self.COLORS:
	    colors.extend(lastColor)
	    colors.extend(color * 2)
	    colors.extend(lastColor)
	    lastColor = color
	return colors

    def _get_triangle_vlist_vertices(self):
	radius = self.INNER_RADIUS - self.BORDER
	centerX = self.center_x
	centerY = self.center_y
	cosA = math.cos(self.tri_angle + math.pi * 2 / 3)
	sinA = math.sin(self.tri_angle + math.pi * 2 / 3)
	cosB = math.cos(self.tri_angle)
	sinB = math.sin(self.tri_angle)
	cosC = math.cos(self.tri_angle + math.pi * 4 / 3)
	sinC = math.sin(self.tri_angle + math.pi * 4 / 3)
	xA = centerX + cosA * self.INNER_RADIUS
	yA = centerY + sinA * self.INNER_RADIUS
	xB = centerX + cosB * self.INNER_RADIUS
	yB = centerY + sinB * self.INNER_RADIUS
	xC = centerX + cosC * self.INNER_RADIUS
	yC = centerY + sinC * self.INNER_RADIUS
	self.point_a = [xA, yA]
	self.vector_i = [xB - xA, yB - yA]
	self.vector_j = [xC - xA, yC - yA]
	return [xB, yB,
		xA, yA,
		xC, yC,
		centerX + cosB * radius, centerY + sinB * radius,
		centerX + cosA * radius, centerY + sinA * radius,
		centerX + cosC * radius, centerY + sinC * radius]

    def _get_triangle_vlist_colors(self):
	pi_3 = math.pi / 3
	angle = self.tri_angle
	if angle < 0:
	    angle += math.pi * 2
	fract_angle = math.fmod(angle, pi_3)
	sect = int(angle / pi_3)
	fract255 = int(fract_angle * 255 / pi_3)
	if sect < 1:           #   0 to 60  degrees, red -> red+green
	    color = [255, fract255, 0]
	elif sect < 2:         #  60 to 120 degrees, red+green -> green
	    color = [255 - fract255, 255, 0]
	elif sect < 3:         # 120 to 180 degrees, green -> green+blue
	    color = [0, 255, fract255]
	elif sect < 4:         # 180 to 240 degrees, green+blue -> blue
	    color = [0, 255 - fract255, 255]
	elif sect < 5:         # 240 to 300 degrees, blue -> blue+red
	    color = [fract255, 0, 255]
	else:                   # 300 to 360 degrees, blue+red -> red
	    color = [255, 0, 255 - fract255]
	self.tip_color = color
	return [255, 255, 255] * 3 + color + [0, 0, 0] + [255, 255, 255]

    def _update_crosshair(self, set_color=True):
	x, y = self.point_a
	xI, yI = self.vector_i
	xJ, yJ = self.vector_j
	if self.crosshair is not None:
	    width, height = self.crosshair.get_needed_size(0, 0)
	    self.crosshair.update(
		x + xI * self.pointer_i + xJ * self.pointer_j - width/2,
		y + yI * self.pointer_i + yJ * self.pointer_j - width/2,
		width, height)
	rI, gI, bI = self.tip_color
	if set_color:
	    self.color = [
		int(min(self.pointer_i * rI + self.pointer_j * 255, 255)),
		int(min(self.pointer_i * gI + self.pointer_j * 255, 255)),
		int(min(self.pointer_i * bI + self.pointer_j * 255, 255)),
		int(self.alpha)]
	    if self.on_select:
		if self.id is not None:
		    self.on_select(id, self.color)
		else:
		    self.on_select(self.color)
	if self.inner_circle_vlist is not None:
	    self.inner_circle_vlist.colors = \
		self.color * (self.NUM_SEGMENTS * 3)

    def _update_triangle(self):
	if self.triangle_vlist is not None:
	    self.triangle_vlist.vertices = self._get_triangle_vlist_vertices()
	    self.triangle_vlist.colors = self._get_triangle_vlist_colors()

    def delete(self):
	if self.circle_vlist is not None:
	    self.circle_vlist.delete()
	    self.circle_vlist = None
	if self.inner_circle_vlist is not None:
	    self.inner_circle_vlist.delete()
	    self.inner_circle_vlist = None
	if self.inner_circle_bg_vlist is not None:
	    self.inner_circle_bg_vlist.delete()
	    self.inner_circle_bg_vlist = None
	if self.colors_vlist is not None:
	    self.colors_vlist.delete()
	    self.colors_vlist = None
	if self.triangle_vlist is not None:
	    self.triangle_vlist.delete()
	    self.triangle_vlist = None
	if self.crosshair is not None:
	    self.crosshair.delete()
	    self.crosshair = None

    def hit_test(self, x, y):
	dX = x - self.center_x
	dY = y - self.center_y
	radius = math.sqrt(dX * dX + dY * dY)
	if radius <= (self.RADIUS + self.BORDER):
	    return True

    def layout(self, x, y):
	Control.layout(self, x, y)
	radius = self.RADIUS + self.BORDER
	self.center_x = self.x + radius
	self.center_y = self.y + radius

	if self.circle_vlist is not None:
	    self.circle_vlist.vertices = self._get_circle_vlist_vertices()
	if self.inner_circle_vlist is not None:
	    self.inner_circle_vlist.vertices = \
		self._get_inner_circle_vlist_vertices()
	if self.inner_circle_bg_vlist is not None:
	    self.inner_circle_bg_vlist.vertices = \
		self._get_inner_circle_vlist_vertices()
	if self.colors_vlist is not None:
	    self.colors_vlist.vertices = self._get_colors_vlist_vertices()
	if self.triangle_vlist is not None:
	    self.triangle_vlist.vertices = self._get_triangle_vlist_vertices()
	    self.triangle_vlist.colors = self._get_triangle_vlist_colors()

    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
	if self.is_turning:
	    self.pointer_i = 1
	    self.pointer_j = 0
	    dX = x - self.center_x
	    dY = y - self.center_y
	    self.tri_angle = math.atan2(dY, dX)
	    self._update_triangle()
	elif self.is_dragging:
	    ax, ay = self.point_a
	    x = x - ax
	    y = y - ay
	    a, b = self.vector_i
	    c, d = self.vector_j
	    i = min(max((d * x - c * y) / (d * a - b * c), 0), 1)
	    j = min(max((a * y - b * x) / (a * d - b * c), 0), 1)
	    if (i + j) > 1.0:
		s = 1.0 / (i + j)
		i = i * s
		j = j * s
	    self.pointer_i = i
	    self.pointer_j = j
	self._update_crosshair()

    def on_mouse_press(self, x, y, button, modifiers):
	dX = x - self.center_x
	dY = y - self.center_y
	radius = math.sqrt(dX * dX + dY * dY)
	if radius >= self.INNER_RADIUS:
	    self.is_turning = True
	    self.is_dragging = False
	else:
	    self.is_dragging = True
	    self.is_turning = False
	self.on_mouse_drag(x, y, 0, 0, button, modifiers)

    def on_mouse_release(self, x, y, button, modifiers):
	self.is_turning = False
	self.is_dragging = False

    def set_alpha(self, alpha):
	self.alpha = alpha
	self.color = self.color[0:3] + [alpha]
	self._update_triangle()
	self._update_crosshair()

    def set_color(self, color):
	r, g, b, a = color
	self.alpha = a
	x, y, z = r, g, b

	if r == g and g == b:  # If it's plain grayscale, any angle works
	    angle = 0
	    i = 0
	    j = float(r) / 255.0
	else:
	    # We normalize the color to its extreme, i.e. 128, 64, 64 would
	    # be normalized to 255, 128, 128 to show the color that would
	    # pertain at the tip
	    highest = max(r, max(g, b))
	    if highest != 255:
		r = r * 255 / highest
		g = g * 255 / highest
		b = b * 255 / highest

	    # Now determine the triangle angle and tip color
	    sector = math.pi / 3
	    if r == 255:
		if g > b:   #   0-60:  r = 255,   g = 0-255, b = 0
		    angle = sector * g / 255
		    b = 0
		    j = float(z) / 255.0
		    x = x - z
		    y = y - z
		    z = 0
		    if x == 255:  # red is maxed, use green to calculate
			i = float(y) / float(g)
		    else:
			i = float(x) / float(r)
		else:       # 300-360: r = 255,   g = 0,     b = 255-0
		    angle = sector * 5 + sector * (255 - b) / 255
		    g = 0
		    j = float(y) / 255.0
		    x = x - y
		    z = z - y
		    y = 0
		    if x == 255:  # red is maxed, use blue to calculate
			if z == 0:  # oops, they're both zero.  Psych!
			    i = 1.0
			else:
			    i = float(z) / float(b)
		    else:
			i = float(x) / float(r)
	    elif g == 255:
		if r > b:   #  60-120: r = 255-0, g = 255,   b = 0
		    angle = sector + sector * (255 - r) / 255
		    b = 0
		    j = float(z) / 255.0
		    x = x - z
		    y = y - z
		    z = 0
		    if y == 255:  # green is maxed, use red to calculate
			i = float(x) / float(r)
		    else:
			i = float(y) / float(g)
		else:       # 120-180: r = 0,     g = 255,   b = 0-255
		    angle = sector * 2 + sector * b / 255
		    r = 0
		    j = float(x) / 255.0
		    y = y - x
		    z = z - x
		    x = 0
		    if y == 255:  # green is maxed, use blue to calculate
			if z == 0:  # oops, both red and blue are zero
			    i = 1.0
			else:
			    i = float(z) / float(b)
		    else:
			i = float(y) / float(g)
	    elif b == 255:
		if g > r:   # 180-240: r = 0,     g = 255-0, b = 255
		    angle = sector * 3 + sector * (255 - g) / 255
		    r = 0
		    j = float(x) / 255.0
		    y = y - x
		    z = z - x
		    x = 0
		    if z == 255:  # blue is maxed, use green to calculate
			i = float(y) / float(g)
		    else:
			i = float(z) / float(b)
		else:       # 240-300: r = 0-255, g = 0,     b = 255
		    angle = sector * 4 + sector * r / 255
		    g = 0
		    j = float(y) / 255.0
		    x = x - y
		    z = z - y
		    y = 0
		    if z == 255:  # blue is maxed, use red to calculate
			if x == 0:  # oops, both green and red are zero
			    i = 1.0
			else:
			    i = float(x) / float(r)
		    else:
			i = float(z) / float(b)

	self.pointer_i = i
	self.pointer_j = j
	self.tri_angle = angle
	self.color = color
	self._update_triangle()
	self._update_crosshair(set_color=False)

    def size(self, dialog):
	Control.size(self, dialog)
	self.width = 2 * (self.RADIUS + self.BORDER)
	self.height = self.width
	if self.circle_vlist is None:
	    self.circle_vlist = dialog.batch.add(4 * self.NUM_SEGMENTS,
		gl.GL_QUADS,
                dialog.fg_group,
                ('v2f', self._get_circle_vlist_vertices()),
                ('c4B', [255, 255, 255, 255] * (4 * self.NUM_SEGMENTS)))
	if self.inner_circle_vlist is None:
	    self.inner_circle_vlist = dialog.batch.add(3 * self.NUM_SEGMENTS,
		gl.GL_TRIANGLES,
		dialog.fg_group,
		('v2f', self._get_inner_circle_vlist_vertices()),
		('c4B', self.color * (3 * self.NUM_SEGMENTS)))
	if self.inner_circle_bg_vlist is None:
	    self.inner_circle_bg_vlist = dialog.batch.add(
		3 * self.NUM_SEGMENTS,
		gl.GL_TRIANGLES,
		dialog.bg_group,
		('v2f', self._get_inner_circle_vlist_vertices()),
		('c4B', self._get_inner_circle_bg_vlist_colors()))
	if self.colors_vlist is None:
	    self.colors_vlist = dialog.batch.add(4 * self.NUM_SEGMENTS,
		gl.GL_QUADS,
                dialog.fg_group,
                ('v2f', self._get_colors_vlist_vertices()),
                ('c3B', self._get_colors_vlist_colors()))
	if self.triangle_vlist is None:
	    self.triangle_vlist = dialog.batch.add(6, gl.GL_TRIANGLES,
		dialog.highlight_group,
		('v2f', self._get_triangle_vlist_vertices()),
		('c3B', self._get_triangle_vlist_colors()))
	if self.crosshair is None:
	    self.crosshair = dialog.theme[self.path]['image'].generate(
                [255, 255, 255, 255],
                dialog.batch,
                dialog.highlight_group)

class ColorSelector(Control):
    """
    Depicts a small color-filled box.  When clicked upon, creates a pop-up
    dialog which can be used to set the color.
    """
    path = ['colorselector', 'swatch']

    def __init__(self, color, width=16, height=16,
		 id=None, on_select=None):
	Control.__init__(self, id=id)
	self.color = color
	self.content_width = width
	self.content_height = height
	self.swatch_label = None
	self.swatch = None
	self.vlist = None
	self.on_select = on_select
	self.select_dialog = None
	self.wheel = None
	self.red_input = None
	self.green_input = None
	self.blue_input = None
	self.slider = None
	self.accept_button = None
	self.cancel_button = None

    def _delete_select_dialog(self):
        if self.select_dialog is not None:
            self.select_dialog.window.remove_handlers(self.select_dialog)
            self.select_dialog.teardown()
            self.select_dialog = None
	    self.wheel = None
	    self.red_input = None
	    self.green_input = None
	    self.blue_input = None
	    self.slider = None
	    self.accept_button = None
	    self.cancel_button = None

    def _get_vlist_vertices(self):
	if self.swatch is None:
	    return [0, 0, 0, 0, 0, 0, 0, 0]
	x, y, width, height = self.swatch.get_content_region()
	return [x, y, x + width, y, x + width, y + height, x, y + height]

    def delete(self):
	Control.delete(self)
	if self.swatch_label is not None:
	    self.swatch_label.delete()
	    self.swatch_label = None
	if self.swatch is not None:
	    self.swatch.delete()
	    self.swatch = None  # deleted as part of layout
	if self.vlist is not None:
	    self.vlist.delete()
	    self.vlist = None
	self._delete_select_dialog()

    def layout(self, x, y):
	Control.layout(self, x, y)
	if self.swatch is not None:
	    width, height = self.swatch.get_needed_size(
		self.content_width, self.content_height)
	    self.swatch.update(self.x, self.y, width, height)
	    if self.vlist is not None:
		self.vlist.vertices = self._get_vlist_vertices()
	if self.swatch_label is not None:
	    self.swatch_label.layout(
		x + self.swatch.width + 4,
		y + (self.height - self.swatch_label.height) / 2)

    def on_mouse_release(self, x, y, button, modifiers):
        if self.is_disabled():
            return

        if self.select_dialog is not None:
            self._delete_select_dialog()  # if it's already up, close it
            return

        # Setup some callbacks for the dialog
        root = self.saved_dialog.get_root()

        def on_escape(dialog):
            self._delete_select_dialog()

        def on_color_set(color):
            self.color = color
	    if self.red_input is not None:
		self.red_input.set_text(str(color[0]))
	    if self.green_input is not None:
		self.green_input.set_text(str(color[1]))
	    if self.blue_input is not None:
		self.blue_input.set_text(str(color[2]))

	def on_red_set(red):
	    red = min(max(int(red), 0), 255)
	    print "red = %s" % red
	    self.color = [red] + self.color[1:]
	    self.wheel.set_color(self.color)

	def on_green_set(green):
	    green = min(max(int(green), 0), 255)
	    print "green = %s" % green
	    self.color = [self.color[0], green] + self.color[2:]
	    self.wheel.set_color(self.color)

	def on_blue_set(blue):
	    blue = min(max(int(blue), 0), 255)
	    print "blue = %s" % blue
	    self.color = self.color[:2] + [blue, self.color[3]]
	    self.wheel.set_color(self.color)

	def on_alpha_set(alpha):
	    self.wheel.set_alpha(int(alpha))

	def on_set_color_button():
	    if self.on_select is not None:
		if self.id is not None:
		    self.on_select(self.id, repr(self.color))
		else:
		    self.on_select(repr(self.color))
            if self.vlist is not None:
                self.vlist.delete()
                self.vlist = None
	    if self.swatch_label is not None:
		self.swatch_label.set_text(repr(self.color))
            self._delete_select_dialog()
            self.saved_dialog.set_needs_layout()

	def on_cancel_button():
            self._delete_select_dialog()
            self.saved_dialog.set_needs_layout()

        # We'll need the root window to get window size
        width, height = root.window.get_size()

        # Now to setup the dialog
	self.wheel = ColorWheel(self.color, on_select=on_color_set)
	self.slider = Slider(value=self.wheel.alpha,
			     min_value=0.0, max_value=255.0,
			     steps=8, width=256, on_set=on_alpha_set)
	self.accept_button = Button("Set Color", on_click=on_set_color_button)
	self.cancel_button = Button("Cancel", on_click=on_cancel_button)
	self.red_input = Input(text=str(self.color[0]), on_input=on_red_set,
			       length=3, max_length=3)
	self.green_input = Input(text=str(self.color[1]),
				 on_input=on_green_set, length=3, max_length=3)
	self.blue_input = Input(text=str(self.color[2]), on_input=on_blue_set,
				length=3, max_length=3)
        self.select_dialog = Dialog(
            Frame(
		VerticalLayout([
		    self.wheel,
		    HorizontalLayout([
			Label("Red"), self.red_input, None,
			Label("Green"), self.green_input, None,
			Label("Blue"), self.blue_input
		    ]),
		    HorizontalLayout([
			Label("Alpha"),
			self.slider,
		    ]),
		    HorizontalLayout([
			self.accept_button, None, self.cancel_button
		    ]),
		])
            ),
            window=root.window, batch=root.batch,
            group=root.root_group.parent, theme=root.theme,
            movable=True, anchor=ANCHOR_CENTER,
            on_escape=on_escape)
        root.window.push_handlers(self.select_dialog)

    def size(self, dialog):
	Control.size(self, dialog)
	if self.swatch is None:
	    self.swatch = dialog.theme[self.path]['image'].generate(
                dialog.theme[self.path]['gui_color'],
                dialog.batch,
                dialog.fg_group)
	if self.vlist is None:
	    self.vlist = dialog.batch.add(4, gl.GL_QUADS,
                dialog.highlight_group,
                ('v2i', self._get_vlist_vertices()),
                ('c4B', self.color * 4))
	if self.swatch_label is None:
	    self.swatch_label = Label(repr(self.color))
	self.swatch_label.size(dialog)
	swatch_width, swatch_height = self.swatch.get_needed_size(
	    self.content_width, self.content_height)
	self.height = max(self.swatch_label.height, swatch_height)
	self.width = swatch_width + 4 + self.swatch_label.width

    def teardown(self):
	Control.teardown(self)
	if self.select_dialog is not None:
	    self.select_dialog.teardown()
	    self.select_dialog = None


# -*- coding: utf-8 -*-
"""setup -- setuptools setup file for kytten.
GUI Framework for Pyglet
"""

__author__ = "Conrad 'Lynx' Wong(originator) Raymond Chandler III(forker)"
__author_email__ = "raymondchandleriii@gmail.com"
__version__ = "6.0.0"
__date__ = "2011 07 31"

try:
    import setuptools
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()

from setuptools import setup, find_packages

f = open('README','rU')
long_description = f.read()
f.close()

setup(
    name = "kytten",
    version = __version__,
    author = "Conrad 'Lynx' Wong(originator) Raymond Chandler III(fork)",
    license="BSD",
    description = "GUI Framework for Pyglet",
    long_description=long_description,
    url = "https://github.com/kitanata/kytten",
    download_url = "https://github.com/kitanata/kytten/tarball/master",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: MacOS X",
        "Environment :: Win32 (MS Windows)",
        "Environment :: X11 Applications",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        ("Topic :: Software Development :: Libraries :: Python Modules"),
        ("Topic :: Games/Entertainment"),
        ],
 
    packages = find_packages(),

    install_requires=['pyglet>=1.1.4'],

    dependency_links=['http://code.google.com/p/pyglet/downloads/list'],

    include_package_data = True,
    zip_safe = False,
)



# proxy_dialog.py
# Copyrighted (C) 2009 by Conrad "Lynx" Wong

import pyglet
import kytten

class ProxyDialog(kytten.Wrapper):
    """
    Allows us to insert the theme being worked on into our own dialog.
    """
    def __init__(self, content, theme):
	kytten.Wrapper.__init__(self, content)
	self.content = content
	self.theme = theme
	self.batch = None
        self.root_group = None
        self.panel_group = None
        self.bg_group = None
        self.fg_group = None
        self.highlight_group = None

    def delete(self):
	kytten.Wrapper.delete(self)
	self.root_group = None
        self.panel_group = None
        self.bg_group = None
        self.fg_group = None
        self.highlight_group = None

    def get_root(self):
        if self.saved_dialog:
            return self.saved_dialog.get_root()
        else:
            return self

    def set_needs_layout(self):
        if self.saved_dialog is not None:
            self.saved_dialog.set_needs_layout()

    def size(self, dialog):
        if dialog is None:
            return
        kytten.Widget.size(self, dialog)
        if self.root_group is None: # do we need to re-clone dialog groups?
            self.batch = dialog.batch
            self.root_group = dialog.fg_group
            self.panel_group = pyglet.graphics.OrderedGroup(
                0, self.root_group)
            self.bg_group = pyglet.graphics.OrderedGroup(
                1, self.root_group)
            self.fg_group = pyglet.graphics.OrderedGroup(
                2, self.root_group)
            self.highlight_group = pyglet.graphics.OrderedGroup(
                3, self.root_group)
            kytten.Wrapper.delete(self)  # rebuild children
        kytten.Wrapper.size(self, self)  # all children are to use our groups



# kytten/widgets.py
# Copyrighted (C) 2009 by Conrad "Lynx" Wong

# Simple widgets belong in this file, to avoid cluttering the directory with
# many small files.  More complex widgets should be placed in separate files.

# Widget: the base GUI element.  A fixed area of space.
# Control: a Widget which accepts events.
# Test: a Widget which draws a crossed box within its area.
# Spacer: a Widget which can expand to fill available space.  Useful to
#         push other Widgets in a layout to the far right or bottom.
# Graphic: a Widget with a texture drawn over its surface.  Can be expanded.
# Label: a Widget which wraps a simple text label.

import pyglet
from pyglet import gl
from override import KyttenLabel

class Widget:
    """
    The base of all Kytten GUI elements.  Widgets correspond to areas on the
    screen and may (in the form of Controls) respond to user input.
    A simple Widget can be used as a fixed-area spacer.

    Widgets are constructed in two passes: first, they are created and
    passed into a Dialog, or added to a Dialog or one of its children,
    then the Dialog calls their size() method to get their size and their
    layout() method to place them on the screen.  When their size is gotten
    for the first time, they initialize any requisite graphic elements
    that could not be done at creation time.
    """
    def __init__(self, width=0, height=0):
        """
        Creates a new Widget.

        @param width Initial width
        @param height Initial height
        """
        self.x = self.y = 0
        self.width = width
        self.height = height
        self.saved_dialog = None

    def _get_controls(self):
        """
        Return this widget if it is a Control, or any children which
        are Controls.
        """
        return []

    def delete(self):
        """
        Deletes any graphic elements we have constructed.  Note that
        we may be asked to recreate them later.
        """
        pass

    def ensure_visible(self):
        if self.saved_dialog is not None:
            self.saved_dialog.ensure_visible(self)

    def expand(self, width, height):
        """
        Expands the widget to fill the specified space given.

        @param width Available width
        @param height Available height
        """
        assert False, "Widget does not support expand"

    def hit_test(self, x, y):
        """
        True if the given point lies within our area.

        @param x X coordinate of point
        @param y Y coordinate of point
        @returns True if the point is within our area
        """
        return x >= self.x and x < self.x + self.width and \
               y >= self.y and y < self.y + self.height

    def is_expandable(self):
        """
        Returns true if the widget can expand to fill available space.
        """
        return False

    def is_focusable(self):
        """
        Return true if the widget can be tabbed to and accepts keyboard
        input
        """
        return False

    def is_input(self):
        """
        Returns true if the widget accepts an input and can return a value
        """
        return False

    def layout(self, x, y):
        """
        Assigns a new location to this widget.

        @param x X coordinate of our lower left corner
        @param y Y coordinate of our lower left corner
        """
        self.x, self.y = x, y

    def size(self, dialog):
        """
        Constructs any graphic elements needed, and recalculates our size
        if necessary.

        @param dialog The Dialog which contains this Widget
        """
        if dialog != self and dialog is not None:
            self.saved_dialog = dialog

    def teardown(self):
        """
        Removes all resources and pointers to other GUI widgets.
        """
        self.delete()
        self.saved_dialog = None

class Control(Widget, pyglet.event.EventDispatcher):
    """
    Controls are widgets which can accept events.

    Dialogs will search their children for a list of controls, and will
    then dispatch events to whichever control is currently the focus of
    the user's attention.
    """
    def __init__(self, id=None, value=None, width=0, height=0, disabled=False):
        """
        Creates a new Control.

        @param id Controls may have ids, which can be used to identify
                  them to the outside application.
        @param value Controls may be assigned values at start time.  The
                     values of all controls which have ids can be obtained
                     through the containing Dialog.
        @param x Initial X coordinate of lower left corner
        @param y Initial Y coordinate of lower left corner
        @param width Initial width
        @param height Initial height
        @param disabled True if control should be disabled
        """
        Widget.__init__(self, width, height)
        self.id = id
        self.value = value
        self.disabled_flag = disabled
        pyglet.event.EventDispatcher.__init__(self)
        self.highlight_flag = False
        self.focus_flag = False

    def _get_controls(self):
        return [(self, self.x, self.x + self.width,    # control, left, right,
                       self.y + self.height, self.y)]  # top, bottom

    def disable(self):
        self.disabled_flag = True
        self.delete()
        if self.saved_dialog is not None:
            self.saved_dialog.set_needs_layout()

    def enable(self):
        self.disabled_flag = False
        self.delete()
        if self.saved_dialog is not None:
            self.saved_dialog.set_needs_layout()

    def get_cursor(self, x, y):
        return self.cursor

    def is_disabled(self):
        return self.disabled_flag

    def is_focus(self):
        return self.focus_flag

    def is_highlight(self):
        return self.highlight_flag

    def on_gain_focus(self):
        self.focus_flag = True

    def on_gain_highlight(self):
        self.highlight_flag = True

    def on_lose_focus(self):
        self.focus_flag = False

    def on_lose_highlight(self):
        self.highlight_flag = False

# Controls can potentially accept most of the events defined for the window,
# but in practice we'll only pass selected events from Dialog.  This avoids
# a large number of unsightly empty method declarations.
for event_type in pyglet.window.Window.event_types:
    Control.register_event_type(event_type)

Control.register_event_type('on_gain_focus')
Control.register_event_type('on_gain_highlight')
Control.register_event_type('on_lose_focus')
Control.register_event_type('on_lose_highlight')
Control.register_event_type('on_update')

class Spacer(Widget):
    """
    A Spacer is an empty widget that expands to fill space in layouts.
    Use Widget if you need a fixed-sized spacer.
    """
    def __init__(self, width=0, height=0):
        """
        Creates a new Spacer.  The width and height given are the minimum
        area that we must cover.

        @param width Minimum width
        @param height Minimum height
        """
        Widget.__init__(self)
        self.min_width, self.min_height = width, height

    def expand(self, width, height):
        """
        Expand the spacer to fill the maximum space.

        @param width Available width
        @param height Available height
        """
        self.width, self.height = width, height

    def is_expandable(self):
        """Indicates the Spacer can be expanded"""
        return True

    def size(self, dialog):
        """Spacer shrinks down to the minimum size for placement.

        @param dialog Dialog which contains us"""
        if dialog is None:
            return
        Widget.size(self, dialog)
        self.width, self.height = self.min_width, self.min_height

class Graphic(Widget):
    """
    Lays out a graphic from the theme, i.e. part of a title bar.
    """
    def __init__(self, path, is_expandable=False):
        Widget.__init__(self)
        self.path = path
        self.expandable=is_expandable
        self.graphic = None
        self.min_width = self.min_height = 0

    def delete(self):
        if self.graphic is not None:
            self.graphic.delete()
            self.graphic = None

    def expand(self, width, height):
        if self.expandable:
            self.width, self.height = width, height
            self.graphic.update(self.x, self.y, self.width, self.height)

    def is_expandable(self):
        return self.expandable

    def layout(self, x, y):
        self.x, self.y = x, y
        self.graphic.update(x, y, self.width, self.height)

    def size(self, dialog):
        if dialog is None:
            return
        Widget.size(self, dialog)
        if self.graphic is None:
            template = dialog.theme[self.path]['image']
            self.graphic = template.generate(
                dialog.theme[self.path]['gui_color'],
                dialog.batch,
                dialog.fg_group)
            self.min_width = self.graphic.width
            self.min_height = self.graphic.height
        self.width, self.height = self.min_width, self.min_height

class Label(Widget):
    """A wrapper around a simple text label."""
    def __init__(self, text="", bold=False, italic=False,
                 font_name=None, font_size=None, color=None, path=[]):
        Widget.__init__(self)
        self.text = text
        self.bold = bold
        self.italic = italic
        self.font_name = font_name
        self.font_size = font_size
        self.color = color
        self.path = path
        self.label = None

    def delete(self):
        if self.label is not None:
            self.label.delete()
            self.label = None

    def layout(self, x, y):
        Widget.layout(self, x, y)
        font = self.label.document.get_font()
        self.label.x = x
        self.label.y = y - font.descent

    def set_text(self, text):
        self.text = text
        self.delete()
        if self.saved_dialog is not None:
            self.saved_dialog.set_needs_layout()

    def size(self, dialog):
        if dialog is None:
            return
        Widget.size(self, dialog)
        if self.label is None:
            self.label = KyttenLabel(
                self.text, bold=self.bold, italic=self.italic,
                color=self.color or
                    dialog.theme[self.path + ['gui_color']],
                font_name=self.font_name or
                          dialog.theme[self.path + ['font']],
                font_size=self.font_size or
                          dialog.theme[self.path + ['font_size']],
                batch=dialog.batch, group=dialog.fg_group)
            font = self.label.document.get_font()
            self.width = self.label.content_width
            self.height = font.ascent - font.descent  # descent is negative




# kytten/theme.py
# Copyrighted (C) 2009 by Conrad "Lynx" Wong

import os

import pyglet
from pyglet import gl

try:
    import json
    json_load = json.loads
except ImportError:
    try:
	import simplejson as json
	json_load = json.loads
    except ImportError:
	import sys
	print >>sys.stderr, \
	      "Warning: using 'safe_eval' to process json files, " \
	      "please upgrade to Python 2.6 or install simplejson"
	import safe_eval
	def json_load(expr):
	    # strip carriage returns
	    return safe_eval.safe_eval(''.join(str(expr).split('\r')))

DEFAULT_THEME_SETTINGS = {
    "font": "Lucida Grande",
    "font_size": 12,
    "font_size_small": 10,
    "text_color": [255, 255, 255, 255],
    "gui_color": [255, 255, 255, 255],
    "highlight_color": [255, 255, 255, 64],
    "disabled_color": [160, 160, 160, 255],
}

class ThemeTextureGroup(pyglet.graphics.TextureGroup):
    """
    ThemeTextureGroup, in addition to setting the texture, also ensures that
    we map to the nearest texel instead of trying to interpolate from nearby
    texels.  This prevents 'blooming' along the edges.
    """
    def set_state(self):
	pyglet.graphics.TextureGroup.set_state(self)
	gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER,
			   gl.GL_NEAREST)
	gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER,
			   gl.GL_NEAREST)

class UndefinedGraphicElementTemplate:
    def __init__(self, theme):
	self.theme = theme
	self.width = 0
	self.height = 0
	self.margins = [0, 0, 0, 0]
	self.padding = [0, 0, 0, 0]

    def generate(self, color, batch, group):
	return UndefinedGraphicElement(self.theme, color, batch, group)

    def write(self, f, indent=0):
        f.write('None')

class TextureGraphicElementTemplate(UndefinedGraphicElementTemplate):
    def __init__(self, theme, texture, width=None, height=None):
	UndefinedGraphicElementTemplate.__init__(self, theme)
	self.texture = texture
	self.width = width or texture.width
	self.height = height or texture.height

    def generate(self, color, batch, group):
	return TextureGraphicElement(self.theme, self.texture,
				     color, batch, group)

    def write(self, f, indent=0):
	f.write('{\n')
	f.write(' ' * (indent + 2) + '"src": "%s"' % self.texture.src)
	if hasattr(self.texture, 'region'):
	    f.write(',\n' + ' ' * (indent + 2) + '"region": %s' %
		    repr(list(self.texture.region)))
	f.write('\n' + ' ' * indent + '}')

class FrameTextureGraphicElementTemplate(TextureGraphicElementTemplate):
    def __init__(self, theme, texture, stretch, padding,
		 width=None, height=None):
	TextureGraphicElementTemplate.__init__(self, theme, texture,
					       width=width, height=height)
	self.stretch_texture = texture.get_region(*stretch).get_texture()
	x, y, width, height = stretch
	self.margins = (x, texture.width - width - x,   # left, right
			texture.height - height - y, y) # top, bottom
	self.padding = padding

    def generate(self, color, batch, group):
	return FrameTextureGraphicElement(
	    self.theme, self.texture, self.stretch_texture,
	    self.margins, self.padding, color, batch, group)

    def write(self, f, indent=0):
	f.write('{\n')
	f.write(' ' * (indent + 2) + '"src": "%s"' % self.texture.src)
	if hasattr(self.texture, 'region'):
	    f.write(',\n' + ' ' * (indent + 2) + '"region": %s' %
		    repr(list(self.texture.region)))
	left, right, top, bottom = self.margins
	if left != 0 or right != 0 or top != 0 or bottom != 0 or \
	   self.padding != [0, 0, 0, 0]:
	    stretch = [left, bottom,
		       self.width - right - left, self.height - top - bottom]
	    f.write(',\n' + ' ' * (indent + 2) + '"stretch": %s' %
		    repr(list(stretch)))
	    f.write(',\n' + ' ' * (indent + 2) + '"padding": %s' %
		    repr(list(self.padding)))
	f.write('\n' + ' ' * indent + '}')

class TextureGraphicElement:
    def __init__(self, theme, texture, color, batch, group):
	self.x = self.y = 0
	self.width, self.height = texture.width, texture.height
	self.group = ThemeTextureGroup(texture, group)
	self.vertex_list = batch.add(4, gl.GL_QUADS, self.group,
				     ('v2i', self._get_vertices()),
				     ('c4B', color * 4),
				     ('t3f', texture.tex_coords))

    def _get_vertices(self):
	x1, y1 = int(self.x), int(self.y)
	x2, y2 = x1 + int(self.width), y1 + int(self.height)
	return (x1, y1, x2, y1, x2, y2, x1, y2)

    def delete(self):
	self.vertex_list.delete()
	self.vertex_list = None
	self.group = None

    def get_content_region(self):
	return (self.x, self.y, self.width, self.height)

    def get_content_size(self, width, height):
	return width, height

    def get_needed_size(self, content_width, content_height):
	return content_width, content_height

    def update(self, x, y, width, height):
	self.x, self.y, self.width, self.height = x, y, width, height
	if self.vertex_list is not None:
	    self.vertex_list.vertices = self._get_vertices()

class FrameTextureGraphicElement:
    def __init__(self, theme, texture, inner_texture, margins, padding,
		 color, batch, group):
	self.x = self.y = 0
	self.width, self.height = texture.width, texture.height
	self.group = ThemeTextureGroup(texture, group)
	self.outer_texture = texture
	self.inner_texture = inner_texture
	self.margins = margins
	self.padding = padding
	self.vertex_list = batch.add(36, gl.GL_QUADS, self.group,
				     ('v2i', self._get_vertices()),
				     ('c4B', color * 36),
				     ('t2f', self._get_tex_coords()))

    def _get_tex_coords(self):
        x1, y1 = self.outer_texture.tex_coords[0:2] # outer's lower left
        x4, y4 = self.outer_texture.tex_coords[6:8] # outer's upper right
        x2, y2 = self.inner_texture.tex_coords[0:2] # inner's lower left
        x3, y3 = self.inner_texture.tex_coords[6:8] # inner's upper right
        return (x1, y1, x2, y1, x2, y2, x1, y2,  # bottom left
                x2, y1, x3, y1, x3, y2, x2, y2,  # bottom
                x3, y1, x4, y1, x4, y2, x3, y2,  # bottom right
                x1, y2, x2, y2, x2, y3, x1, y3,  # left
                x2, y2, x3, y2, x3, y3, x2, y3,  # center
                x3, y2, x4, y2, x4, y3, x3, y3,  # right
                x1, y3, x2, y3, x2, y4, x1, y4,  # top left
                x2, y3, x3, y3, x3, y4, x2, y4,  # top
                x3, y3, x4, y3, x4, y4, x3, y4)  # top right

    def _get_vertices(self):
        left, right, top, bottom = self.margins
        x1, y1 = int(self.x), int(self.y)
        x2, y2 = x1 + int(left), y1 + int(bottom)
        x3 = x1 + int(self.width) - int(right)
	y3 = y1 + int(self.height) - int(top)
        x4, y4 = x1 + int(self.width), y1 + int(self.height)
        return (x1, y1, x2, y1, x2, y2, x1, y2,  # bottom left
                x2, y1, x3, y1, x3, y2, x2, y2,  # bottom
                x3, y1, x4, y1, x4, y2, x3, y2,  # bottom right
                x1, y2, x2, y2, x2, y3, x1, y3,  # left
                x2, y2, x3, y2, x3, y3, x2, y3,  # center
                x3, y2, x4, y2, x4, y3, x3, y3,  # right
                x1, y3, x2, y3, x2, y4, x1, y4,  # top left
                x2, y3, x3, y3, x3, y4, x2, y4,  # top
                x3, y3, x4, y3, x4, y4, x3, y4)  # top right

    def get_content_region(self):
	left, right, top, bottom = self.padding
	return (self.x + left, self.y + bottom,
		self.width - left - right, self.height - top - bottom)

    def get_content_size(self, width, height):
	left, right, top, bottom = self.padding
	return width - left - right, height - top - bottom

    def get_needed_size(self, content_width, content_height):
	left, right, top, bottom = self.padding
	return (max(content_width + left + right, self.outer_texture.width),
	        max(content_height + top + bottom, self.outer_texture.height))

    def delete(self):
	self.vertex_list.delete()
	self.vertex_list = None
	self.group = None

    def update(self, x, y, width, height):
	self.x, self.y, self.width, self.height = x, y, width, height
	if self.vertex_list is not None:
	    self.vertex_list.vertices = self._get_vertices()

class UndefinedGraphicElement(TextureGraphicElement):
    def __init__(self, theme, color, batch, group):
	self.x = self.y = self.width = self.height = 0
	self.group = group
	self.vertex_list = batch.add(12, gl.GL_LINES, self.group,
				     ('v2i', self._get_vertices()),
				     ('c4B', color * 12))

    def _get_vertices(self):
	x1, y1 = int(self.x), int(self.y)
	x2, y2 = x1 + int(self.width), y1 + int(self.height)
	return (x1, y1, x2, y1, x2, y1, x2, y2,
		x2, y2, x1, y2, x1, y2, x1, y1,
		x1, y1, x2, y2, x1, y2, x2, y1)

class ScopedDict(dict):
    """
    ScopedDicts differ in several useful ways from normal dictionaries.

    First, they are 'scoped' - if a key exists in a parent ScopedDict but
    not in the child ScopedDict, we return the parent value when asked for it.

    Second, we can use paths for keys, so we could do this:
        path = ['button', 'down', 'highlight']
	color = theme[path]['highlight_color']

    This would return the highlight color assigned to the highlight a button
    should have when it is clicked.
    """
    def __init__(self, arg={}, parent=None):
	self.parent = parent
	for k, v in arg.iteritems():
	    if isinstance(v, dict):
		self[k] = ScopedDict(v, self)
	    else:
		self[k] = v

    def __getitem__(self, key):
	if key is None:
	    return self
	elif isinstance(key, list) or isinstance(key, tuple):
	    if len(key) > 1:
		return self.__getitem__(key[0]).__getitem__(key[1:])
	    elif len(key) == 1:
		return self.__getitem__(key[0])
	    else:
		return self  # theme[][key] should return theme[key]
	else:
	    try:
		return dict.__getitem__(self, key)
	    except KeyError:
		if self.parent is not None:
		    return self.parent.__getitem__(key)
		else:
		    raise

    def __setitem__(self, key, value):
	if isinstance(value, dict):
	    dict.__setitem__(self, key, ScopedDict(value, self))
	else:
	    dict.__setitem__(self, key, value)

    def get(self, key, default=None):
	if isinstance(key, list) or isinstance(key, tuple):
	    if len(key) > 1:
		return self.__getitem__(key[0]).get(key[1:], default)
	    elif len(key) == 1:
		return self.get(key[0], default)
	    else:
		raise KeyError(key)  # empty list

	if self.has_key(key):
	    return dict.get(self, key)
	elif self.parent:
	    return self.parent.get(key, default)
	else:
	    return default

    def get_path(self, path, default=None):
	assert isinstance(path, list) or isinstance(path, tuple)
	if len(path) == 1:
	    return self.get(path[0], default)
	else:
	    return self.__getitem__(path[0]).get_path(path[1:], default)

    def set_path(self, path, value):
	assert isinstance(path, list) or isinstance(path, tuple)
	if len(path) == 1:
	    return self.__setitem__(path[0], value)
	else:
	    return self.__getitem__(path[0]).set_path(path[1:], value)

    def write(self, f, indent=0):
	f.write('{\n')
	first = True
	for k, v in self.iteritems():
	    if not first:
		f.write(',\n')
	    else:
		first = False
	    f.write(' ' * (indent + 2) + '"%s": ' % k)
	    if isinstance(v, ScopedDict):
		v.write(f, indent + 2)
	    elif isinstance(v, UndefinedGraphicElementTemplate):
		v.write(f, indent + 2)
	    elif isinstance(v, basestring):
		f.write('"%s"' % v)
	    elif isinstance(v, tuple):
		f.write('%s' % repr(list(v)))
	    else:
		f.write(repr(v))
	f.write('\n')
	f.write(' ' * indent + '}')

class Theme(ScopedDict):
    """
    Theme is a dictionary-based class that converts any elements beginning
    with 'image' into a GraphicElementTemplate.  This allows us to specify
    both simple textures and 9-patch textures, and more complex elements.
    """
    def __init__(self, arg, override={}, default=DEFAULT_THEME_SETTINGS,
		 allow_empty_theme=False, name='theme.json'):
	"""
	Creates a new Theme.

	@param arg The initializer for Theme.  May be:
	    * another Theme - we'll use the same graphic library but
	                      apply an override for its dictionary.
	    * a dictionary - interpret any subdirectories where the key
			     begins with 'image' as a GraphicElementTemplate
	    * a filename - read the JSON file as a dictionary
	@param override Replace some dictionary entries with these
	@param default Initial dictionary entries before handling input
	@param allow_empty_theme True if we should allow creating a new theme
	"""
	ScopedDict.__init__(self, default, None)

	self.groups = {}

	if isinstance(arg, Theme):
	    self.textures = arg.textures
	    for k, v in arg.iteritems():
		self.__setitem__(k, v)
	    self.update(override)
	    return

	if isinstance(arg, dict):
	    self.loader = pyglet.resource.Loader(os.getcwd())
	    input = arg
	else:
	    if os.path.isfile(arg) or os.path.isdir(arg):
		self.loader = pyglet.resource.Loader(path=arg)
		try:
		    theme_file = self.loader.file(name)
		    input = json_load(theme_file.read())
		    theme_file.close()
		except pyglet.resource.ResourceNotFoundException:
		    input = {}
	    else:
		input = {}

	self.textures = {}
	self._update_with_images(self, input)
	self.update(override)

    def __getitem__(self, key):
	try:
	    return ScopedDict.__getitem__(self, key)
	except KeyError, e:
	    if key.startswith('image'):
		return UndefinedGraphicElementTemplate(self)
	    else:
		raise e

    def _get_texture(self, filename):
	"""
	Returns the texture associated with a filename.  Loads it from
	resources if we haven't previously fetched it.

	@param filename The filename of the texture
	"""
	if not self.textures.has_key(filename):
	    texture = self.loader.texture(filename)
	    texture.src = filename
	    self.textures[filename] = texture
	return self.textures[filename]

    def _get_texture_region(self, filename, x, y, width, height):
	"""
	Returns a texture region.

	@param filename The filename of the texture
	@param x X coordinate of lower left corner of region
	@param y Y coordinate of lower left corner of region
	@param width Width of region
	@param height Height of region
	"""
	texture = self._get_texture(filename)
	retval = texture.get_region(x, y, width, height).get_texture()
	retval.src = texture.src
	retval.region = [x, y, width, height]
	return retval

    def _update_with_images(self, target, input):
	"""
	Update a ScopedDict with the input dictionary.  Translate
	images into texture templates.

	@param target The ScopedDict which is to be populated
	@param input The input dictionary
	"""
	for k, v in input.iteritems():
	    if k.startswith('image'):
		if isinstance(v, dict):
		    width = height = None
		    if v.has_key('region'):
			x, y, width, height = v['region']
			texture = self._get_texture_region(
				v['src'], x, y, width, height)
		    else:
			texture = self._get_texture(v['src'])
		    if v.has_key('stretch'):
			target[k] = FrameTextureGraphicElementTemplate(
			    self,
			    texture,
			    v['stretch'],
			    v.get('padding', [0, 0, 0, 0]),
			width=width, height=height)
		    else:
			target[k] = TextureGraphicElementTemplate(
			    self, texture, width=width, height=height)
		else:
		    target[k] = TextureGraphicElementTemplate(
			self, self._get_texture(v))
	    elif isinstance(v, dict):
		temp = ScopedDict(parent=target)
		self._update_with_images(temp, v)
		target[k] = temp
	    else:
		target[k] = v

    def write(self, f, indent=0):
	ScopedDict.write(self, f, indent)
	f.write('\n')
